#!/usr/bin/env python
""" cdecl.py - parse c declarations

(c) 2002, 2003 Simon Burton <simon@arrowtheory.com>
Released under GNU GPL license.

version 0.xx

"""
from __future__ import generators

import sys
import string
import types
import copy

# ==================================================== #

class Lexer:

	def __init__(self,s="",verbose=0,**kw):
		self.verbose = verbose
		# map keywords and typedefs
		self.lookup = {}
		for t in \
			"float double void char int".split():
			self.lookup[t] = BasicType( t )
		for t in \
			"register signed unsigned short long const volatile".split():
			self.lookup[t] = Qualifier( t )
		for t in "extern static auto".split():
			self.lookup[t] = StorageClass( t )
		self.lookup['typedef'] = Typedef()
		#self.lookup['__inline__'] = GCCBuiltin('__inline__')
		self.lookup['...'] = Ellipses()
		if s:
			self.lex(s)
		for key in kw.keys():
			self.__dict__[key] = kw[key]

	def lex(self,s):
		self.stack = None
		self.lines = s.splitlines()
		self.set_state("","",0,0)
		self.so_file = ""
		self._newline()
		self.get_token() # start

	def mktypedef(self,tok,node):
		if self.verbose:
			print "%s.mktypedef(%s,%s)"%(self,tok,node)
		self.lookup[ tok ] = node

	def rmtypedef(self,tok):
		" used in round trip testing "
		print "# rmtypedef(%s)"%tok
		assert isinstance( self.lookup[ tok ], Node ) # existance
		del self.lookup[ tok ]

	def _get_kind(self,tok):
		#print '_get_kind(%s)'%tok,self.lookup
		try:
			return self.lookup[tok]
			return self.lookup[tok].clone()
		except KeyError:
			if tok.startswith("__builtin"):
				node = GCCBuiltin(tok)
				self.lookup[tok] = node
				return node
			return None

	def _newline(self):
		while self.lno < len(self.lines):
			line = self.lines[self.lno]
			if not line or line[0] != "#":
				break
			l = line.split('"')
			assert len(l)>=2
			self.so_file = l[1]
			#self.so_lno = int( l[0].split()[1] )
			#sys.stderr.write("# %s %s: %s\n"%(so_lno,so_file,l))
			self.lno+=1

	def get_token(self):
		self.push_state()
		ident_chars0 = string.letters+"_"
		ident_chars1 = string.letters+string.digits+"_"
		tok, kind = "", ""
		while self.lno < len(self.lines):
			s = self.lines[self.lno]
			i=self.col
			while i < len(s):
				if s[i].isspace():
					i=i+1
					continue
				#if s[i] in ident_chars0:
				if s[i].isalpha() or s[i]=='_':
					# identifier
					j=i+1
					while j<len(s):
						if s[j] in ident_chars1:
							j=j+1
						else:
							break
					tok = s[i:j]
					self.col = j
					kind = self._get_kind(tok)
					break
				if s[i].isdigit() or \
					(i+1<len(s) and s[i] in '+-.' and s[i+1].isdigit()):
					# number literal
					is_float = s[i]=='.'
					is_hex = s[i:i+2]=='0x'
					if is_hex:
						i=i+2
						assert s[i].isdigit()
					j=i+1
					while j<len(s):
						#print "lex ",repr(s[i]),is_float
						if s[j].isdigit():
							j=j+1
						elif s[j]=='.' and not is_float:
							assert not is_hex
							j=j+1
							is_float=1
						else:
							break 
					tok = s[i:j]
					self.col = j
					if is_float:
						kind = float(tok)
					elif is_hex:
						kind = int(tok,16)
					else:
						kind = int(tok)
					break
				if s[i:i+3]=='...':
					# ellipses
					#sys.stderr.write( "ELLIPSES "+str(self.get_state()) )
					tok = s[i:i+3]
					kind = self._get_kind(tok)
					self.col = i+3
					break
				if s[i] in '*{}()[]:;,=+-.<>|&':
					tok = s[i]
					kind = tok
					self.col = i+1
					break
				# keep moving
				#sys.stderr.write( "lexer ignoring '%s'\n"%s[i] )
				i=i+1
			if i==len(s):
				# nothing found
				assert tok == ""
				self.col=0
				self.lno+=1
				self._newline()
			else:
				assert tok
				break
		self.set_state(tok,kind,self.lno,self.col)

	def err_string(self):
		"Return helpful error string :)"
		return self.lines[self.lno]+"\n"+" "*self.col+"^\n"

	def push_state(self):
		self.stack = self.get_state() # short stack
		#self.stack.push( self.get_state() )

	def unget_token(self):
		assert self.stack is not None
		self.set_state(*self.stack)
		self.stack = None

	def set_state(self,tok,kind,lno,col):
		if self.verbose:
			print "tok,kind,lno,col = ",(tok,kind,lno,col)
		self.tok = tok
		self.kind = kind
		self.lno = lno # line
		self.col = col # column

	def get_state(self):
		return self.tok,self.kind,self.lno,self.col

	def get_file(self):
		return self.so_file

# ==================================================== #

class Symbols:

	def __init__(self,parent=None):
		self.verbose = verbose
		self.parent=parent # are we a nested namespace?
		self.lookup = {} # identifiers
		self.tags = {} # struct, union, enum tags

	def __str__(self):
		return "Symbols(%s,%s)"%(self.lookup,self.tags)

	def __getitem__(self,key):
		try:
			item = self.lookup[key]
		except KeyError:
			item = None
			#if self.parent is not None:
				#item = self.parent[item]
				## self[key] = item # cache
		#if self.verbose: print "%s.get('%s')='%s'"%(self,key,item)
		return item

	def __setitem__(self,key,val):
		#if self.verbose: print "%s.set('%s','%s')"%(self,key,val)
		assert val is not None
		self.lookup[key] = val

	def set_tag(self,key,val):
		#if self.verbose: print "%s.set_tag(%s,%s)"%(self,key,val)
		assert len(key)
		self.tags[key] = val

	def deep_get_tag(self,key):
		try:
			item = self.tags[key]
		except KeyError:
			item = None
			if self.parent is not None:
				item = self.parent.deep_get_tag(key)
		#if self.verbose: print "%s.get_tag(%s)=%s"%(self,key,item)
		return item

	def get_tag(self,key):
		try:
			item = self.tags[key]
		except KeyError:
			item = None
		#if self.verbose: print "%s.get_tag(%s)=%s"%(self,key,item)
		return item

# ==================================================== #

class ParseError(Exception):

	def __init__(self,*e):
		self.e = e

	def __str__(self):
		return string.join(map(str,self.e))

class Parser:

	def parse_error(self,lexer,reason,*blah):
		sys.stderr.write( "%s.parse_error()\n"%self.deepstr() )
		sys.stderr.write( "at line %s: %s\n"%(lexer.lno+1,reason) )
		sys.stderr.write( lexer.err_string() )
		raise ParseError(reason,*blah)

	def expected_error(self,lexer,*l):
		self.parse_error( lexer, "expected %s, got '%s'"\
			%(string.join(map(repr,l)," or "),lexer.tok))

	def consume(self,lexer,tok):
		if lexer.tok != tok:
			self.expected_error(lexer, tok)
		lexer.get_token()

	def parse_enter(self,lexer):
		#return
		self.start_lno=lexer.lno
		self.file=lexer.so_file

	def parse_leave(self,lexer):
		#return
		self.lines =\
			lexer.lines[self.start_lno:max(lexer.lno,self.start_lno+1)]

# ==================================================== #

class Node(list):
	" A node in a parse tree "

	def __init__(self,*items,**kw):
		list.__init__( self, items )
		self.lock1 = 0
		self.lock2 = 0
		self.verbose = 0
		for key in kw.keys():
			self.__dict__[key] = kw[key]

	def __str__(self):
		return "%s%s"%(self.__class__.__name__,tuple(self))

	def __repr__(self):
		return str(self)

	def __eq__(self,other):
		if not isinstance(other,Node):
			return 0
		if len(self)!=len(other):
			return 0
		for i in range(len(self)):
			if not self[i]==other[i]:
				return 0
		return 1

	def __ne__(self,other):
		return not self==other

	def filter(self,klass):
		return [x for x in self if isinstance(x,klass)]
		#return filter( lambda x:isinstance(x,klass), self )

	def find(self,klass):
		for x in self:
			if isinstance(x,klass):
				return x
		return None

	def deep_find(self,klass):
		for x in self:
			if isinstance(x,Node):
				if isinstance(x,klass):
					return x
				node = x.deep_find(klass)
				if node is not None:
					return node
		return None

	def leaves(self):
		for i in self:
			if isinstance( i, Node ):
				for j in i.leaves():
					yield j
			else:
				yield i

	def nodes(self):
		for i in self:
			if isinstance( i, Node ):
				for j in i.nodes():
					yield j
		yield self

	def deeplen(self):
		i=0
		if not self.lock2:
			self.lock2=1
			for item in self:
				i+=1
				if isinstance(item,Node):
					i+=item.deeplen()
			self.lock2=0
		else:
			i+=1
		return i

	def deepstr(self,level=0):
		if self.deeplen() < 4:
			nl = ""; indent = ""
		else:
			nl="\n"; indent = "  "
		s = []
		if not self.lock1:
			self.lock1=1
			for item in self:
				if isinstance(item,Node):
					s.append( indent*(level+1)+item.deepstr(level+1) )
				else:
					s.append( indent*(level+1)+repr(item) )
			self.lock1=0
		else:
			for item in self:
				if isinstance(item,Node):
					s.append( indent*(level+1)+"..." )
				else:
					s.append( indent*(level+1)+"%s"%repr(item) )
		return "%s(%s)"%\
			(self.__class__.__name__,nl+string.join(s,","+nl))

	def clone(self):
		" a slow clone "
		return copy.deepcopy(self)

	def clone(self):
		l = []
		for item in self:
			if isinstance(item,Node):
				item = item.clone()
			l.append(item)
		# we skip any attributes...
		return self.__class__(*l)

	def expose(self,klass):
		' expose children of any <klass> instance '
		# children first
		for x in self:
			if isinstance(x,Node):
				x.expose(klass)
		# now the tricky bit
		i=0
		while i < len(self):
			if isinstance(self[i],klass):
				node=self.pop(i)
				for x in node:
					assert not isinstance(x,klass)
					# pass on some attributes
					if hasattr(node,'lines') and not hasattr(x,'lines'):
						x.lines=node.lines
					if hasattr(node,'file') and not hasattr(x,'file'):
						x.file=node.file
					self.insert(i,x) # expose
					i=i+1
					assert i<=len(self)
			else:
				i=i+1

	def delete(self,klass):
		' delete any <klass> subtree '
		for x in self:
			if isinstance(x,Node):
				x.delete(klass)
		# now the tricky bit
		i=0
		while i < len(self):
			if isinstance(self[i],klass):
				self.pop(i)
			else:
				i=i+1

	def idem(self,klass):
		" <klass> is made idempotent "
		# children first
		for x in self:
			if isinstance(x,Node):
				x.idem(klass)
		if isinstance(self,klass):
			# now the tricky bit
			i=0
			while i < len(self):
				if isinstance(self[i],klass):
					node = self.pop(i)
					for x in node:
						assert not isinstance(x,klass)
						self.insert(i,x) # idempotent
						i=i+1
						assert i<=len(self)
				else:
					i=i+1

	def is_typedef(self):
		for x in self:
			if isinstance(x,Node):
				if x.is_typedef():
					return 1
		return 0

	def is_struct(self):
		for x in self:
			if isinstance(x,Node):
				if x.is_struct():
					return 1
		return 0

	def get_declarator(self):
		return Declarator( Identifier(), self[:] )

	#def explain(self):
		#l = []
		#for x in self:
			#if isinstance(x,Node):
				#l.append(x.explain())
			#else:
				#l.append(str(x))
		#return string.join(l," ") 
			##(self.__class__.__name__,string.join(l) )

	def psource(self):
		if hasattr(self,'lines'):
			print "# "+string.join(self.lines,"\n# ")+"\n"

	def report(self):
		self.psource()
		print self.deepstr()

	def cstr(self,l=None):
		#print '%s.cstr(%s)'%(self,l)
		if l is None:
			l = []
		for x in self:
			if isinstance(x,Node):
				x.cstr(l)
			else:
				l.insert(0,str(x)+' ')
		return string.join(l,'')

	def ctype(self):
		#print "%s.ctype()"%self
		l=[]
		for x in self:
			if isinstance(x,Node):
				l.append(x.ctype())
			else:
				l.append(x)
		#print "%s.__class__(*%s)"%(self,l)
		return self.__class__(*l)

# ==================================================== #

class BasicType(Node):
	" int char short etc. "
	def __init__(self,name):
		Node.__init__(self,name)

class Qualifier(Node):
	def __init__(self,name):
		Node.__init__(self,name)
		self.name=name

class StorageClass(Node):
	def __init__(self,name):
		Node.__init__(self,name)
		self.name=name

class Typedef(StorageClass):
	def __init__(self,s='typedef'):
		Node.__init__(self,s)
	#def explain(self):
		#return "type"

class Ellipses(Node):
	def __init__(self,s='...'):
		Node.__init__(self,s)

class GCCBuiltin(BasicType):
	pass

class Identifier(Node):

	def __init__(self,name="",*items):
		if name or 1:
			Node.__init__(self,name,*items)
		else:
			Node.__init__(self)
		self.name=name

	#def explain(self):
		#if len(self)==1:
			#return "%s"%self.name
		#else:
			#return "%s initialized to %s"%(self.name,
				#Node(self[1]).explain()) # will handle Initializer

	def ctype(self):
		#assert len(self)==2
		return self.__class__(*self[1:]) #.clone() ?

	def cstr(self,l=None):
		if l is None:
			l=[]
		if len(self)>1:
			assert len(self)==2
			l.append( '%s = %s'%(self[0],self[1]) )
		else:
			l.append( str(self[0]) )
		return string.join(l)

class Function(Node,Parser):

	def __init__(self,*items):
		Node.__init__(self,*items)

	def parse(self,lexer,symbols):
		symbols = Symbols(symbols)
		args = ''
		#lexer.get_token()
		if lexer.tok != ')':
			if not lexer.tok: self.parse_error(lexer)
			#lexer.unget_token() # unget start of decl
			while lexer.tok != ')':
				node = ParameterDeclaration()
				node.parse(lexer,symbols)
				self.append( node )
				if lexer.tok != ')' and lexer.tok != ',':
					self.parse_error(lexer)
				if lexer.tok == ',':
					lexer.get_token()
		lexer.get_token()

	#def explain(self):
		#if len(self):
			#return "function (%s), returning"%\
				#string.join( map(lambda x:x.explain(),self), ", " )
		#else:
			#return "function returning"

	def is_typedef(self):
		return self[-1].is_typedef()

	def cstr(self,l):
		#print '%s.cstr(%s)'%(self,l)
		_l=[]
		assert len(self)
		i=0
		while isinstance(self[i],Declarator):
			_l.append( self[i].cstr() )
			i=i+1
		l.append( '(%s)'%string.join(_l, ',') )
		while i<len(self):
			self[i].cstr(l)
			i=i+1
		return string.join(l)

	def return_type(self):
		node = self[-1]
		#assert isinstance(node,DeclarationSpecifiers)
		return Declarator( node )

	def arg_types(self):
		return [arg.ctype() for arg in self[:-1]]

	def ctype(self):
		return Function(*self.arg_types()+[self[-1]])

class Pointer(Node):

	def __init__(self,*items):
		Node.__init__(self,*items)

	#def explain(self):
		#return "pointer to"

	def cstr(self,l):
		assert len(self)
		node=self[0]
		l.insert(0,'*')
		if isinstance(node,Function):
			l.insert(0,'(')
			l.append(')')
		elif isinstance(node,Array):
			l.insert(0,'(')
			l.append(')')
		return Node.cstr(self,l)

class Array(Node,Parser):

	def __init__(self,*items):
		Node.__init__(self,*items)

	def parse(self,lexer,symbols):
		lexer.get_token() # a number or ']'
		# XXX hack hack
		while lexer.tok != ']':
			#self.append( lexer.kind )
			self.append( lexer.tok )
			lexer.get_token()
		lexer.get_token() # read past the ']'

	#def explain(self):
		#s=''
		#if len(self):
			#if type(self[0])==int:
				#s='0 to %s '%(self[0]-1)
		#return "array %sof"%s

	def cstr(self,l):
		if len(self)==1:
			l.append('[]')
		elif len(self)==2:
			l.append('[%s]'%self[1])
		else:
			# expression in array size
			# assert 0
			l.append('[')
			for tok in self[1:]: l.append(tok)
			l.append(']')
		return self[0].cstr(l)

class Tag(Node):
	pass

class Compound(Node,Parser):
	"Struct or Union"

	def __init__(self,*items,**kw):
		Node.__init__(self,*items,**kw)

	def parse(self,lexer,symbols):
		symbols = Symbols(symbols)
		tag = "" # anonymous
		if lexer.tok != '{':
			tag = lexer.tok
			if not ( tag[0]=='_' or tag[0].isalpha() ):
				self.parse_error(lexer ,"expected tag, got '%s'"%tag )
			lexer.get_token()
		if tag:
			self.append(Tag(tag))
		else:
			self.append(Tag())
		self.tag = tag
		if lexer.tok == '{':
			fieldlist = []
			lexer.get_token()
			if lexer.tok != '}':
				if not lexer.tok: self.parse_error(lexer)
				while lexer.tok != '}':
					node = StructDeclaration()
					node.parse(lexer,symbols)
					fieldlist.append( node )
			self += fieldlist
			lexer.get_token()
		if self.verbose: 
			print "%s.__init__() #<--"%(self)

	def cstr(self,_l=None):
		assert isinstance( self[0], Tag )
		tag=''
		if len(self[0]):
			tag=' '+self[0][0]
		if isinstance(self,Struct):
			l=[ 'struct%s '%tag ]
		elif isinstance(self,Union):
			l=[ 'union%s '%tag ]
		if len(self)>1:
			l.append('\n{\n')
			for decl in self[1:]:
				l.append( decl.cstr()+";\n" )
			l.append('} ')
		if _l is None:
			_l=[]
		while l:
			_l.insert( 0, l.pop() )
		return string.join( _l, '' )

	def ctype(self):
		if len(self[0])==0:
			#print "# WARNING : anonymous struct " # OK i think
			return self.clone()
		return self.__class__( self[0] ) # just the Tag

	def has_members(self):
		return self.find( StructDeclarator ) is not None

class Struct(Compound):

	#def explain(self):
		#tag=self.tag
		#if len(tag): tag=' '+tag
		#if len(self)>1:
			#return "struct%s containing %s"%\
				#( tag,
				#string.join( map(lambda x:x.explain(),self[1:])," and " ) )
		#else:
			#return "struct%s"%tag
	pass

class Union(Compound):

	#def explain(self):
		#if len(self)>1:
			#return "union %s containing %s"%\
				#( self.tag,
				#string.join( map(lambda x:x.explain(),self[1:])," and " ) )
		#else:
			#return "union %s"%self.tag
	pass

class Enum(Node,Parser):

	def __init__(self,*items,**kw):
		Node.__init__(self,*items,**kw)

	def parse(self,lexer,symbols):
		tag = "" # anonymous
		if lexer.tok != '{':
			tag = lexer.tok
			if not ( tag[0]=='_' or tag[0].isalpha() ):
				self.parse_error(lexer ,"expected tag, got '%s'"%tag )
			lexer.get_token()
		if tag:
			self.append(Tag(tag))
		else:
			self.append(Tag())
		self.tag = tag
		if lexer.tok == '{':
			lexer.get_token()
			if lexer.tok != '}':
				if not lexer.tok:
					self.parse_error(lexer)
				while lexer.tok != '}':
					if lexer.kind is not None:
						self.expected_error(lexer ,"identifier" )
					ident = Identifier(lexer.tok)
					if symbols[ident[0]] is not None:
						self.parse_error(lexer,"%s already defined."%ident[0])
					symbols[ident[0]]=ident
					self.append( ident )
					lexer.get_token()
					if lexer.tok == '=':
						lexer.get_token()
						# ConstantExpr
						if type( lexer.kind ) is not int:
							#self.parse_error(lexer ,"expected integer" )
							# XXX hack hack
							while lexer.tok!=',' and lexer.tok!='}':
								lexer.get_token()
						else:
							# put initializer into the Identifier
							ident.append( lexer.kind )
							lexer.get_token()
					if lexer.tok != '}':
						if lexer.tok != ',':
							self.expected_error(lexer,"}",",")
						lexer.get_token() # ','
			lexer.get_token()
		if self.verbose:
			print "%s.__init__() #<--"%(self)

	#def explain(self):
		#tag = self.find(Tag)
		#tag_s=""
		#if len(tag):
			#tag_s = " "+tag[0]
		#if len(self)>1:
			#return "enumeration%s of %s"%\
				#( tag_s,
				#string.join( map(lambda x:x.explain(),self[1:]),", then " ) )
		#else:
			#return "enumeration%s"%tag_s

	def cstr(self,_l=None):
		assert isinstance( self[0], Tag )
		tag=''
		if len(self[0]):
			tag=' '+self[0][0]
		l=[ 'enum%s '%tag ]
		if len(self)>1:
			l.append('\n{\n')
			for node in self[1:]:
				l.append( node.cstr()+',\n' )
			l.append('} ')
		if _l is None:
			_l=[]
		while l:
			_l.insert( 0, l.pop() )
		return string.join( _l, '' )

	def ctype(self):
		if len(self[0])==0:
			return self.clone()
		return self.__class__( self[0] ) # just the Tag

class Declarator(Node,Parser):

	def __init__(self,*items):
		Node.__init__(self,*items)
		self.ident = None

	def parse(self,lexer,symbols):
		#Parser.parse_enter(self,lexer)
		stack = []
		# read up to identifier, pushing tokens onto stack
		self.ident = self.parse_identifier(lexer,symbols,stack)
		self.name = ''
		if self.ident is not None:
			self.append( self.ident )
			self.name = self.ident.name
		# now read outwards from identifier
		self.parse_declarator(lexer,symbols,stack)
		#Parser.parse_leave(self,lexer)

	def parse_identifier(self,lexer,symbols,stack):
		if self.verbose:
			print "%s.parse_identifier()"%self
		ident = None
		if lexer.tok != ';':
			while lexer.tok and lexer.kind is not None:
				stack.append( (lexer.tok, lexer.kind) )
				lexer.get_token()
			if lexer.tok:
				ident = Identifier( lexer.tok )
				#stack.append( (ident.name, ident) )
				lexer.get_token()
		if self.verbose:
			print "%s.parse_identifier()=%s"%(self,repr(ident))
		return ident

	def parse_declarator(self,lexer,symbols,stack,level=0):
		if self.verbose:
			print "  "*level+"%s.parse_declarator(%s) # --->"%\
				(self,stack)
		if lexer.tok == '[':
			while lexer.tok == '[':
				node = Array()
				node.parse(lexer,symbols)
				self.append(node)
			if lexer.tok == '(':
				self.parse_error(lexer ,"array of functions" )
		elif lexer.tok == '(':
			lexer.get_token()
			node = Function()
			node.parse(lexer,symbols)
			self.append( node )
			if lexer.tok == '(':
				self.parse_error(lexer ,"function returns a function" )
			if lexer.tok == '[':
				self.parse_error(lexer ,"function returns an array" )
		while stack:
			tok, kind = stack[-1] # peek
			if tok == '(':
				stack.pop()
				self.consume(lexer,')')
				self.parse_declarator(lexer,symbols,stack,level+1)
			elif tok == '*':
				stack.pop()
				self.append( Pointer() )
			else:
				tok, kind = stack.pop()
				self.append( kind )
		if self.verbose:
			print "  "*level+"%s.parse_declarator(%s) # <---"%\
				(self,stack)

	def act(self):
		return
		# children go first 
		for node in self.nodes():
			if isinstance(node,TypeAlias):
				node.do_type()
		#self.expose(TypeAlias)
		#self.idem(DeclarationSpecifiers)

	def is_struct(self):
		declspec = self.find(DeclarationSpecifiers)
		if declspec is not None:
			struct = declspec.find(Struct)
			if struct is not None:
				return 1
		return 0

	def get_struct(self):
		declspec = self.find(DeclarationSpecifiers)
		if declspec is not None:
			struct = declspec.find(Struct)
			if struct is not None:
				return struct
		return None

	def is_function(self):
		return len(self)>1 and isinstance(self[1],Function)

	def get_function(self):
		if len(self)>1 and isinstance(self[1],Function):
			return self[1]
		return None

	def is_pointer(self):
		return self.find(Pointer) is not None

	def get_pointer_deref(self):
		pointer = self.find(Pointer) 
		return Declarator( Identifier(), pointer[:] )

	def is_typedef(self):
		return self[-1].is_typedef()
		##return self.filter(Typedef)
		#declspec = self.deep_find(DeclarationSpecifiers)
		#assert declspec is not None # did you ask me too soon?
		#return declspec.find(Typedef)
		###return filter( lambda x: x=='typedef', self.leaves() )

	#def explain(self):
		#if len(self):
			#if isinstance( self[0], Identifier ):
				#return "%s is %s"%(self[0].explain(),
					#string.join(map(lambda x:x.explain(),self[1:])," "))
			#else:
				#return Node.explain(self)
		#return ''

	#def rawcstr(self,l=None):
		#return Node.cstr(self,l).strip()

	def cstr(self,l=None):
		return Node.cstr(self,l).strip()

	def ctype(self):
		decl=self.clone()
		if isinstance(decl[0],Identifier):
			decl.pop(0)
		decl.insert(0,Identifier())
		for i in range(1,len(decl)):
			decl[i]=decl[i].ctype()
		return decl

	def cname(self):
		" return the name of the identifier "
		ident = self[0]
		assert isinstance( ident, Identifier ) # o.w. return None ?
		return ident[0]

	def declare_str(self,name):
		" return string declaring name with same type as self "
		tp = self.ctype()
		assert isinstance( tp[0], Identifier )
		#tp.insert(0,Identifier(name))
		#tp[0].append(name)
		tp[0][0]=name
		return tp.cstr()+";"

class AbstractDeclarator(Declarator):
	""" used in ParameterDeclaration; may lack an identifier """

	def parse_identifier(self,lexer,symbols,stack):
		if self.verbose:
			print "%s.parse_identifier()"%self
		ident = None
		while 1:
			if lexer.tok == ';':
				self.parse_error(lexer)
			if lexer.tok == ')':
				break
			if lexer.tok == ',':
				break
			if lexer.tok == '[':
				break
			if lexer.kind is None:
				#print "%s.new identifier"%self
				ident = Identifier( lexer.tok )
				lexer.get_token()
				#stack.append( (ident.name, ident) )
				break
			stack.append( (lexer.tok, lexer.kind) )
			lexer.get_token()
		if self.verbose:
			print "%s.parse_identifier()=%s"%(self,repr(ident))
		return ident

	#def cstr(self,l=None):
		#return Node.cstr(self,l)

class FieldLength(Node):

	#def explain(self):
		#return ""

	def cstr(self,l):
		l.append(':%s'%self[0]) 

class StructDeclarator(Declarator):

	def parse(self,lexer,symbols):
		if lexer.tok != ':':
			Declarator.parse(self,lexer,symbols)
		if lexer.tok == ':':
			lexer.get_token()
			# ConstantExpr
			length = int(lexer.tok)
			#print "length = ",length
			self.append( FieldLength(length) )
			lexer.get_token()

	#def explain(self):
		#flen = self.find(FieldLength)
		#if flen is not None:
			#i = self.index(flen)
			#self.pop(i)
			#s = Declarator.explain(self)
			#self.insert(i,flen)
			#width = flen[0]
			#if width > 0:
				#return s+" bitfield %s wide"%width
			#else:
				#return s+" alignment bitfield"
		#else:
			#return Declarator.explain(self)

class DeclarationSpecifiers(Node,Parser):
	def __init__(self,*items):
		Node.__init__(self,*items)

	def __eq__(self,other):
		" unordered (set/bag) equality "
		if not isinstance(other,Node):
			return 0
		for i in range(len(self)):
			if not self[i] in other:
				return 0
		for i in range(len(other)):
			if not other[i] in self:
				return 0
		return 1

	def parse(self,lexer,symbols):
		self.parse_spec(lexer,symbols)
		self.reverse()

	def parse_spec(self,lexer,symbols):
		typespec = None
		while lexer.tok:
			if isinstance( lexer.kind, TypeAlias ) or\
				isinstance( lexer.kind, BasicType ):
				if typespec is not None:
					self.parse_error(lexer ,"type already specified as %s"\
						%typespec )
				typespec=lexer.kind
				self.append( lexer.kind )
				lexer.get_token()
			elif isinstance( lexer.kind, Qualifier ):
				self.append( lexer.kind )
				lexer.get_token()
			elif isinstance( lexer.kind, StorageClass ):
				self.append( lexer.kind )
				lexer.get_token()
			elif lexer.tok=='struct':
				lexer.get_token()
				self.parse_struct(lexer,symbols)
				break #?
			elif lexer.tok=='union':
				lexer.get_token()
				self.parse_union(lexer,symbols)
				break #?
			elif lexer.tok=='enum':
				lexer.get_token()
				self.parse_enum(lexer,symbols)
				break #?
			elif lexer.kind is None:
				# identifier
				break
			else:
				break

	def parse_struct(self,lexer,symbols):
		if self.verbose:
			print "%s.parse_struct()"%(self)
		node = Struct()
		node.parse(lexer,symbols)
		_node = None
		if node.tag:
			_node = symbols.get_tag( node.tag )
		if _node is not None:
			if not isinstance( _node, Struct ):
				self.parse_error(lexer,"tag defined as wrong kind")
			if len(node)>1:
				if len(_node)>1:
					self.parse_error(lexer,"tag already defined as %s"%_node)
				#symbols.set_tag( node.tag, node )
			#else:
				# refer to the previously defined struct
				##node = _node
				#node = _node.clone()
		if 0:
			# refer to the previously defined struct
			if len(node)==1:
				_node = symbols.deep_get_tag( node.tag )
				if _node is not None:
					node=_node
			# But what about any future reference to the struct ?
		if node.tag:
			symbols.set_tag( node.tag, node )
		self.append( node )

	def parse_union(self,lexer,symbols):
		if self.verbose:
			print "%s.parse_union(%s)"%(self,node)
		node = Union()
		node.parse(lexer,symbols)
		_node = None
		if node.tag:
			_node = symbols.get_tag( node.tag )
		if _node is not None:
			if not isinstance( _node, Union ):
				self.parse_error(lexer,"tag %s defined as wrong kind"%repr(node.tag))
			if len(node)>1:
				if len(_node)>1:
					self.parse_error(lexer,"tag already defined as %s"%_node)
				#symbols.set_tag( node.tag, node )
			#else:
				#node = _node
		#if len(node)==1:
			#_node = symbols.deep_get_tag( node.tag )
			#if _node is not None:
				#node=_node
		if node.tag:
			symbols.set_tag( node.tag, node )
		self.append( node )

	def parse_enum(self,lexer,symbols):
		if self.verbose:
			print "%s.parse_enum(%s)"%(self,node)
		node = Enum()
		node.parse(lexer,symbols)
		_node = None
		if node.tag:
			_node = symbols.get_tag( node.tag )
		if _node is not None:
			if not isinstance( _node, Enum ):
				self.parse_error(lexer,"tag defined as wrong kind")
			if len(node)>1:
				if len(_node)>1:
					self.parse_error(lexer,"tag already defined as %s"%_node)
				#symbols.set_tag( node.tag, node )
			#else:
				#node = _node
		#if len(node)==1:
			#_node = symbols.deep_get_tag( node.tag )
			#if _node is not None:
				#node=_node
		if node.tag:
			symbols.set_tag( node.tag, node )
		self.append( node )

	def is_typedef(self):
		return self.find(Typedef) is not None

	def is_struct(self):
		return self.find(Struct) is not None

	def needs_declarator(self):
		for node in self:
			if isinstance( node, Struct ):
				return False 
			if isinstance( node, Enum ):
				return False 
			if isinstance( node, Union ):
				return False 
		return True

class TypeSpecifiers(DeclarationSpecifiers):
	" used in ParameterDeclaration "

	def parse_spec(self,lexer,symbols):
		typespec = None
		while lexer.tok:
			if isinstance( lexer.kind, TypeAlias ) or\
				isinstance( lexer.kind, BasicType ):
				if typespec is not None:
					self.parse_error(lexer ,"type already specified as %s"\
						%typespec )
				typespec=lexer.kind
				self.append( lexer.kind )
				lexer.get_token()
			elif isinstance( lexer.kind, Qualifier ):
				self.append( lexer.kind )
				lexer.get_token()
			elif isinstance( lexer.kind, StorageClass ):
				self.parse_error(lexer ,"'%s' cannot appear here"%lexer.tok )
			elif lexer.tok=='struct':
				lexer.get_token()
				self.parse_struct(lexer,symbols)
				break #?
			elif lexer.tok=='union':
				lexer.get_token()
				self.parse_union(lexer,symbols)
				break #?
			elif lexer.tok=='enum':
				lexer.get_token()
				self.parse_enum(lexer,symbols)
				break #?
			elif lexer.kind is None:
				# identifier
				break
			else:
				break

class Initializer(Node,Parser):

	def __init__(self,*items):
		Node.__init__(self,*items)

	def parse(self,lexer,symbols):
		self.parse_error(lexer,"not implemented")

class TypeAlias(Node):
	" typedefed things "

	def __init__(self,name,decl=None):
		Node.__init__(self,name)#,decl)
		self.name=name
		self.decl=decl

	#def do_type(self):
		#return
		#assert type(self.pop(0))==str
		#self.expose(Declarator)
		#self.delete(Identifier)
		#self.delete(Typedef)

class Declaration(Node,Parser):

	def __init__(self,*items):
		Node.__init__(self,*items)
		#self.acted=False

	def parse(self,lexer,symbols):
		if not lexer.tok: 
			return
		Parser.parse_enter(self,lexer)
		declspec = DeclarationSpecifiers() 
		declspec.parse(lexer,symbols)
		if len(declspec)==0:
			if lexer.tok == ';':
				lexer.get_token()
				# empty declaration...
				return
			self.parse_error(lexer,
				"expected specifiers, got '%s'"%lexer.tok )
		self.append(declspec)
		while 1:
			decl = Declarator()
			decl.parse(lexer,symbols)
			if len(decl)==0:
				if declspec.needs_declarator():
					self.parse_error(lexer,
						"expected declarator, got '%s'"%lexer.tok )
			self.append(decl)
			ident = decl.ident
			if ident is not None:
				# install symbol
				node = symbols[ident[0]]
				if node is not None:
					# we allow functions to be defined (as same) again
					#print node.deepstr(),'\n', self.deepstr() 
					_node = node.clone()
					_node.delete(Identifier)
					_self = self.clone()
					_self.delete(Identifier)
					if _node != _self:
						self.parse_error(lexer,
							"\n%s\n  already defined as \n%s\n"%\
							(self.deepstr(),node.deepstr()))
				else:
					if self.is_typedef():
						#lexer.mktypedef( ident[0], self )
						tp = TypeAlias(ident[0],decl) 
						lexer.mktypedef( ident[0], tp )
					else:
						symbols[ident[0]] = self
				if lexer.tok == '=':
					# parse initializer
					lexer.get_token()
					init = Initializer()
					init.parse(lexer,symbols)
					ident.append( init ) # as in Enum
			#else: struct, union or enum
			if lexer.tok == ';':
				# no more declarators
				break
			if lexer.tok == '{':
				# ! ahhh function !!!
				sys.stderr.write(
					"WARNING: function body found at line %s\n"%lexer.lno )
				while lexer.tok != '}': 
					lexer.get_token()
				lexer.get_token()
				Parser.parse_leave(self,lexer)
				return
			self.consume(lexer,',')
		self.consume(lexer,';')
		Parser.parse_leave(self,lexer)

	def is_typedef(self):
		spec=self[0]
		assert isinstance(spec,DeclarationSpecifiers)
		return spec.is_typedef()

	def do_spec(self):
		" distribute DeclarationSpecifiers over each Declarator "
		spec=self[0]
		assert isinstance(spec,DeclarationSpecifiers)
		self.pop(0)
		for declarator in self:
			assert isinstance(declarator,Declarator)
			#if isinstance(declarator,DeclarationSpecifiers #huh?
			##for node in spec:
				##declarator.append(node.clone())
			declarator.append(spec)

	def act(self):
		# children go first 
		for node in self.nodes():
			if isinstance(node,Declaration):
				node.do_spec()
		self.expose(Declaration)

	#def explain(self):
		#return string.join([x.explain() for x in self],", ") 
		#return string.join(map(lambda x:x.explain(),self),", ") 

class ParameterDeclaration(Declaration):

	def parse(self,lexer,symbols):
		typespec = TypeSpecifiers()
		typespec.parse(lexer,symbols)
		self.append(typespec)
		decl = AbstractDeclarator()
		decl.parse(lexer,symbols)
		self.append(decl)
		ident = decl.ident
		if ident is not None:
			node = symbols[ident[0]]
			if node is not None:
				self.parse_error(lexer,
					"%s already defined as %s"%(ident,node))
			else:
				symbols[ident[0]] = self

class StructDeclaration(Declaration):

	def parse(self,lexer,symbols):
		if not lexer.tok: 
			return
		declspec = DeclarationSpecifiers() 
		declspec.parse(lexer,symbols)
		self.append(declspec)
		if len(declspec)==0:
			if lexer.tok == ';':
				lexer.get_token()
				# empty declaration...
				return
			self.parse_error(lexer,
				"expected specifiers, got '%s'"%lexer.tok )
		while 1:
			decl = StructDeclarator()
			decl.parse(lexer,symbols)
			if len(decl)==0:
				self.parse_error(lexer,
					"expected declarator, got '%s'"%lexer.tok )
			self.append(decl)
			ident = decl.ident
			if ident is not None:
				node = symbols[ident[0]]
				if node is not None:
					self.parse_error(lexer ,
						"%s already defined as %s"%(ident,node))
				else:
					if declspec.is_typedef():
						self.parse_error(lexer,"typedef in struct or union")
					else:
						symbols[ident[0]] = self
			if lexer.tok == ';':
				break
			self.consume(lexer,',')
		self.consume(lexer,';')

class Rewrite:
	" visit nodes in tree "

	def __init__(self,node=None):
		if node is not None:
			self.visit(node)

	def visit(self,node):
		#print 'visit(%s)'%node
		for _node in node:
			if isinstance(_node,Declarator):
				self.visit_decl(_node)
			elif isinstance(_node,Node):
				self.visit(_node)

	def visit_decl(self,decl):
		assert isinstance(decl,Declarator)
		i=len(decl)
		accum=[] # accumulate nodes
		while i:
			i=i-1
			node=decl.pop(i)
			if isinstance(node,Declarator):
				self.visit_decl(node)
			else:
				self.visit(node)
			if isinstance(node,Pointer):
				node+=accum
				accum=[node]
			elif isinstance(node,Function):
				node+=accum
				accum=[node]
			elif isinstance(node,Array):
				while accum:
					node.insert(0,accum.pop())
				accum=[node]
				# array size (if any) at end
			#elif isinstance(node,Identifier):
				#node+=accum
				#accum=[node]
			else:
				# accumulate
				accum.insert(0,node)
		decl[:]=accum

class TransUnit(Node,Parser):

	def __init__(self,*items,**kw):
		Node.__init__(self,*items,**kw)

	def parse(self,s,verbose=0):
		self.symbols = Symbols()
		self.lexer = Lexer(s,verbose=verbose)
		node = None
		while self.lexer.tok:
			node=Declaration()
			node.parse(self.lexer,self.symbols)
			#sys.stderr.write( "# line %s\n"%self.lexer.lno )
			if node:
				self.append(node)
				#node.psource()
				#print node.deepstr(),'\n'
				#node.act()

	def strip(self,files):
		' leave only declarations in <files> '
		i=0
		while i<len(self):
			if self[i].file in files:
				i=i+1
			else:
				self.pop(i)

	def assert_no_dups(self):
		check={}
		for node in self.nodes():
			assert not check.has_key(id(node))
			check[id(node)]=1

	def act(self, verbose=0, test_parse=0, test_types=0 ):
		rw=Rewrite()
		i=0
		while i < len(self):
			if verbose: print "##"*25
			declaration=self[i]
			if verbose: declaration.psource()
			if verbose: print declaration.deepstr(),'\n'
			assert isinstance(declaration,Declaration)
			#try:
			if verbose: print "# declaration.act()"
			declaration.act()
			#except Exception, e:
				#declaration.report()
				#raise e
			if verbose: print declaration.deepstr(),'\n'
			self[i:i+1] = declaration # expose declarators from declaration
			for j in range(len(declaration)):
				declarator=self[i]
				assert isinstance(declarator,Declarator)
				if verbose: print "# declarator.act()"
				declarator.act()
				if verbose: print declarator.deepstr(),'\n'
				if verbose: print "# rw.visit_decl(declarator)"
				rw.visit_decl(declarator)
				if verbose: print declarator.deepstr(),'\n'
				#print declarator.ctype().deepstr(),'\n'
				#assert declarator.clone() == declarator
				if test_parse:
					# test that parse of cstr gives same answer
					cstr = declarator.cstr()+';\n'
					if verbose: print '# '+cstr.replace('\n','\n# ')
					#print
					if declarator.is_typedef():
						name = declarator[0][0]
						assert type(name)==str
						self.lexer.rmtypedef( name )
					declaration = Declaration()
					self.lexer.lex( cstr )
					#print self.lexer.err_string()
					declaration.parse(  self.lexer, Symbols() ) # use new NS
					#declaration.parse(  Lexer( cstr ), Symbols() )
					declaration.act()
					assert len(declaration)==1
					decl=declaration[0]
					decl.act()
					rw.visit_decl(decl)
					if decl!=declarator:
						if verbose: print "#???????????"
						if verbose: print decl.deepstr(),'\n\n'
						#if verbose: print declaration.deepstr(),'\n\n'
						#assert 0
					elif verbose: print '# OK\n'
				if test_types:
					node = declarator.ctype()
					declare_str= node.declare_str("my_name")
					if verbose: print "# declarator.ctype() "
					if verbose: print node.deepstr(),"\n"
					if verbose: print "#",declare_str.replace('\n','\n# ')
				i=i+1

# ==================================================== #

try:
	import NoModule
	import psyco
	from psyco.classes import *
except ImportError:
	class _psyco:
		def jit(self):      pass
		def bind(self, f):  pass
		def proxy(self, f): return f
	psyco = _psyco()
psyco.bind( Lexer.get_token )
psyco.bind( Node )

verbose = 0
if __name__=="__main__":
	verbose = 0
	if not sys.argv[1:]:
		s = sys.stdin.read()
	if sys.argv[1:]:
		s = sys.argv[1]
		if sys.argv[2:]:
			verbose = int(sys.argv[2])
	if 0:
		import profile
		profile.run('TransUnit(s)','prof.out')
		import pstats
		p=pstats.Stats('prof.out')
		p.strip_dirs().sort_stats(-1).print_stats()
	else:
		node = TransUnit(verbose = 1 )
		node.parse(s)
		node.act(1,1,1)
